"""
This module defines special characters which are used when expressing musical terms.
"""


__author__ = "https://codeberg.org/decoherence"


class IDMusicalChar:

    DIM = chr(176)
    """The 'degree' symbol used for diminished chords."""

    FLAT = chr(9837)
    """The 'b' symbol used for flat notes and chords."""

    HALF_DIM = chr(119209)
    """The 'slashed degree' symbol used for half-diminished chords."""